#!/usr/bin/env python3

dbfile = "~/.local/share/look4bas/basis_sets.db"
default_download_formats = ["Gaussian94"]
list_formats = {
    "default": ["crop", "noelements", "colour"],
    "extra":   ["crop", "elements", "colour"]
}
